"""Dictionary Tests"""
from os import remove
from os.path import join

import pytest

from hdx.utilities.dictandlist import (
    args_to_dict,
    avg_dicts,
    dict_diff,
    dict_of_dicts_add,
    dict_of_lists_add,
    dict_of_sets_add,
    extract_list_from_list_of_dict,
    float_value_convert,
    integer_key_convert,
    integer_value_convert,
    key_value_convert,
    list_distribute_contents,
    list_distribute_contents_simple,
    merge_dictionaries,
    read_list_from_csv,
    write_list_to_csv,
)
from hdx.utilities.path import temp_dir


class TestDictAndList:
    def test_merge_dictionaries(self):
        d1 = {1: 1, 2: 2, 3: 3, 4: {"a": 1, "b": "c"}}
        d2 = {2: 6, 4: 8, 6: 9, 9: {"c": 12, "e": "h"}}
        d3 = {4: {"g": 3}, 5: 7, 8: {3: 12, "k": "b"}}
        with pytest.raises(ValueError):
            merge_dictionaries([d1, d2, d3])
        d2[4] = {8: "8"}
        result = merge_dictionaries([d1, d2, d3])
        assert result == {
            1: 1,
            2: 6,
            3: 3,
            4: {8: "8", "g": 3, "b": "c", "a": 1},
            5: 7,
            6: 9,
            8: {"k": "b", 3: 12},
            9: {"e": "h", "c": 12},
        }
        d1 = {1: 1, 2: 2, 3: 3, 4: ["a", "b", "c"]}
        d2 = {2: 6, 5: 8, 6: 9, 4: ["d", "e"]}
        result = merge_dictionaries([d1, d2])
        assert result == {1: 1, 2: 6, 3: 3, 4: ["d", "e"], 5: 8, 6: 9}
        d1 = {1: 1, 2: 2, 3: 3, 4: ["a", "b", "c"]}
        result = merge_dictionaries([d1, d2], merge_lists=True)
        assert result == {
            1: 1,
            2: 6,
            3: 3,
            4: ["a", "b", "c", "d", "e"],
            5: 8,
            6: 9,
        }

    def test_dict_diff(self):
        d1 = {1: 1, 2: 2, 3: 3, 4: {"a": 1, "b": "c"}}
        d2 = {4: {"a": 1, "b": "c"}, 2: 2, 3: 3, 1: 1}
        diff = dict_diff(d1, d2)
        assert diff == {}
        d2[3] = 4
        diff = dict_diff(d1, d2)
        assert diff == {3: (3, 4)}
        del d2[3]
        diff = dict_diff(d1, d2)
        assert diff == {3: (3, "<KEYNOTFOUND>")}
        d2[3] = 3
        del d1[4]
        diff = dict_diff(d1, d2)
        assert diff == {4: ("<KEYNOTFOUND>", {"a": 1, "b": "c"})}

    def test_dict_of_lists_add(self):
        d = {}
        dict_of_lists_add(d, "a", 1)
        assert d == {"a": [1]}
        dict_of_lists_add(d, 2, "b")
        assert d == {"a": [1], 2: ["b"]}
        dict_of_lists_add(d, "a", 2)
        assert d == {"a": [1, 2], 2: ["b"]}
        dict_of_lists_add(d, 2, "c")
        assert d == {"a": [1, 2], 2: ["b", "c"]}
        dict_of_lists_add(d, 2, "b")
        assert d == {"a": [1, 2], 2: ["b", "c", "b"]}

    def test_dict_of_sets_add(self):
        d = {}
        dict_of_sets_add(d, "a", 1)
        assert d == {"a": {1}}
        dict_of_sets_add(d, 2, "b")
        assert d == {"a": {1}, 2: {"b"}}
        dict_of_sets_add(d, "a", 2)
        assert d == {"a": {1, 2}, 2: {"b"}}
        dict_of_sets_add(d, 2, "c")
        assert d == {"a": {1, 2}, 2: {"b", "c"}}
        dict_of_sets_add(d, 2, "b")
        assert d == {"a": {1, 2}, 2: {"b", "c"}}

    def test_dict_of_dicts_add(self):
        d = {}
        dict_of_dicts_add(d, "a", 1, 3.0)
        assert d == {"a": {1: 3.0}}
        dict_of_dicts_add(d, 2, "b", 5.0)
        assert d == {"a": {1: 3.0}, 2: {"b": 5.0}}
        dict_of_dicts_add(d, "a", 2, 4.5)
        assert d == {"a": {1: 3.0, 2: 4.5}, 2: {"b": 5.0}}
        dict_of_dicts_add(d, 2, "c", 9.3)
        assert d == {"a": {1: 3.0, 2: 4.5}, 2: {"b": 5.0, "c": 9.3}}
        dict_of_dicts_add(d, 2, "b", "abc")
        assert d == {"a": {1: 3.0, 2: 4.5}, 2: {"b": "abc", "c": 9.3}}

    def test_list_distribute_contents_simple(self):
        input_list = [3, 1, 1, 1, 2, 2]
        result = list_distribute_contents_simple(input_list)
        assert result == [1, 2, 3, 1, 2, 1]
        input_list = [
            ("f", 1),
            ("d", 7),
            ("d", 2),
            ("c", 2),
            ("c", 5),
            ("c", 4),
            ("f", 2),
        ]
        result = list_distribute_contents_simple(input_list, lambda x: x[0])
        assert result == [
            ("c", 2),
            ("d", 7),
            ("f", 1),
            ("c", 5),
            ("d", 2),
            ("f", 2),
            ("c", 4),
        ]
        input_list = [
            {"key": "d", "data": 5},
            {"key": "d", "data": 1},
            {"key": "g", "data": 2},
            {"key": "a", "data": 2},
            {"key": "a", "data": 3},
            {"key": "b", "data": 5},
        ]
        result = list_distribute_contents_simple(
            input_list, lambda x: x["key"]
        )
        assert result == [
            {"key": "a", "data": 2},
            {"key": "b", "data": 5},
            {"key": "d", "data": 5},
            {"key": "g", "data": 2},
            {"key": "a", "data": 3},
            {"key": "d", "data": 1},
        ]

    def test_list_distribute_contents(self):
        input_list = [3, 1, 1, 1, 2, 2]
        result = list_distribute_contents(input_list)
        assert result == [1, 2, 1, 2, 1, 3]
        input_list = [
            ("f", 1),
            ("d", 7),
            ("d", 2),
            ("c", 2),
            ("c", 5),
            ("c", 4),
            ("f", 2),
        ]
        result = list_distribute_contents(input_list, lambda x: x[0])
        assert result == [
            ("c", 2),
            ("d", 7),
            ("f", 1),
            ("c", 5),
            ("d", 2),
            ("c", 4),
            ("f", 2),
        ]
        input_list = [
            {"key": "d", "data": 5},
            {"key": "d", "data": 1},
            {"key": "g", "data": 2},
            {"key": "a", "data": 2},
            {"key": "a", "data": 3},
            {"key": "b", "data": 5},
        ]
        result = list_distribute_contents(input_list, lambda x: x["key"])
        assert result == [
            {"key": "a", "data": 2},
            {"key": "d", "data": 5},
            {"key": "b", "data": 5},
            {"key": "a", "data": 3},
            {"key": "d", "data": 1},
            {"key": "g", "data": 2},
        ]

    def test_extract_list_from_list_of_dict(self):
        input_list = [
            {"key": "d", 1: 5},
            {"key": "d", 1: 1},
            {"key": "g", 1: 2},
            {"key": "a", 1: 2},
            {"key": "a", 1: 3},
            {"key": "b", 1: 5},
        ]
        result = extract_list_from_list_of_dict(input_list, "key")
        assert result == ["d", "d", "g", "a", "a", "b"]
        result = extract_list_from_list_of_dict(input_list, 1)
        assert result == [5, 1, 2, 2, 3, 5]

    def test_key_value_convert(self):
        d1 = {1: 1, 2: 1.0, 3: 3, 4: 4}
        assert key_value_convert(d1) == d1
        d1 = {1: 2, 2: 2.0, 3: 5, "la": 4}
        assert key_value_convert(d1, keyfn=int) == {
            1: 2,
            2: 2.0,
            3: 5,
            "la": 4,
        }
        assert key_value_convert(d1, keyfn=int, dropfailedkeys=True) == {
            1: 2,
            2: 2.0,
            3: 5,
        }
        d1 = {1: 2, 2: 2.0, 3: 5, 4: "la"}
        assert key_value_convert(d1, valuefn=int) == {
            1: 2,
            2: 2.0,
            3: 5,
            4: "la",
        }
        assert key_value_convert(d1, valuefn=int, dropfailedvalues=True) == {
            1: 2,
            2: 2.0,
            3: 5,
        }

    def test_integer_key_convert(self):
        d1 = {1: 1, 2: 1.5, 3.5: 3, "4": 4}
        assert integer_key_convert(d1) == {1: 1, 2: 1.5, 3: 3, 4: 4}

    def test_integer_value_convert(self):
        d1 = {1: 1, 2: 1.5, 3: "3", 4: 4}
        assert integer_value_convert(d1) == {1: 1, 2: 1, 3: 3, 4: 4}

    def test_float_value_convert(self):
        d1 = {1: 1, 2: 1.5, 3: "3", 4: 4}
        assert float_value_convert(d1) == {1: 1.0, 2: 1.5, 3: 3.0, 4: 4.0}

    def test_avg_dicts(self):
        d1 = {1: 1, 2: 1.0, 3: 3, 4: 4}
        d2 = {1: 2, 2: 2.0, 3: 5, 4: 4, 7: 3}
        assert avg_dicts(d1, d2) == {1: 1.5, 2: 1.5, 3: 4, 4: 4}
        assert avg_dicts(d1, d2, dropmissing=False) == {
            1: 1.5,
            2: 1.5,
            3: 4,
            4: 4,
            7: 3,
        }

    def test_read_write_list_to_csv(self):
        list_of_tuples = [(1, 2, 3, "a"), (4, 5, 6, "b"), (7, 8, 9, "c")]
        list_of_lists = [[1, 2, 3, "a"], [4, 5, 6, "b"], [7, 8, 9, "c"]]
        with temp_dir(
            "TestReadWriteList",
            delete_on_success=True,
            delete_on_failure=False,
        ) as tempdir:
            filename = "test_read_write_list_to_csv.csv"
            filepath = join(tempdir, filename)
            write_list_to_csv(
                filepath, list_of_lists, headers=["h1", "h2", "h3", "h4"]
            )
            newll = read_list_from_csv(filepath)
            newld = read_list_from_csv(filepath, headers=1, dict_form=True)
            remove(filepath)
            assert newll == [
                ["h1", "h2", "h3", "h4"],
                ["1", "2", "3", "a"],
                ["4", "5", "6", "b"],
                ["7", "8", "9", "c"],
            ]
            assert newld == [
                {"h1": "1", "h2": "2", "h4": "a", "h3": "3"},
                {"h1": "4", "h2": "5", "h4": "b", "h3": "6"},
                {"h1": "7", "h2": "8", "h4": "c", "h3": "9"},
            ]
            write_list_to_csv(
                filepath, list_of_tuples, headers=("h1", "h2", "h3", "h4")
            )
            newll = read_list_from_csv(filepath)
            newld = read_list_from_csv(filepath, headers=1, dict_form=True)
            remove(filepath)
            assert newll == [
                ["h1", "h2", "h3", "h4"],
                ["1", "2", "3", "a"],
                ["4", "5", "6", "b"],
                ["7", "8", "9", "c"],
            ]
            assert newld == [
                {"h1": "1", "h2": "2", "h4": "a", "h3": "3"},
                {"h1": "4", "h2": "5", "h4": "b", "h3": "6"},
                {"h1": "7", "h2": "8", "h4": "c", "h3": "9"},
            ]
            write_list_to_csv(
                filepath, list_of_lists, headers=["h1", "h2", "h3"]
            )
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["h1", "h2", "h3"],
                ["1", "2", "3"],
                ["4", "5", "6"],
                ["7", "8", "9"],
            ]
            write_list_to_csv(
                filepath,
                list_of_lists,
                headers=["h1", "h3", "h4"],
                columns=[1, 3, 4],
            )
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["h1", "h3", "h4"],
                ["1", "3", "a"],
                ["4", "6", "b"],
                ["7", "9", "c"],
            ]

            list_of_lists = [
                ["1", "2", "3", "a"],
                ["4", "5", "6", "b"],
                [7, 8, 9, "c"],
            ]
            write_list_to_csv(filepath, list_of_lists)
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["1", "2", "3", "a"],
                ["4", "5", "6", "b"],
                ["7", "8", "9", "c"],
            ]
            write_list_to_csv(filepath, list_of_lists, headers=2)
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["4", "5", "6", "b"],
                ["7", "8", "9", "c"],
            ]

            list_of_dicts = [
                {"h1": 1, "h2": 2, "h3": 3, "h4": "a"},
                {"h1": 4, "h2": 5, "h3": 6, "h4": "b"},
                {"h1": 7, "h2": 8, "h3": 9, "h4": "c"},
            ]
            write_list_to_csv(
                filepath, list_of_dicts, headers=["h1", "h2", "h3", "h4"]
            )
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["h1", "h2", "h3", "h4"],
                ["1", "2", "3", "a"],
                ["4", "5", "6", "b"],
                ["7", "8", "9", "c"],
            ]
            write_list_to_csv(filepath, list_of_dicts)
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["h1", "h2", "h3", "h4"],
                ["1", "2", "3", "a"],
                ["4", "5", "6", "b"],
                ["7", "8", "9", "c"],
            ]
            write_list_to_csv(filepath, list_of_dicts, headers=2)
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["1", "2", "3", "a"],
                ["4", "5", "6", "b"],
                ["7", "8", "9", "c"],
            ]
            write_list_to_csv(
                filepath,
                list_of_dicts,
                headers=["h1", "h3", "h4"],
                columns=["h1", "h3", "h4"],
            )
            newll = read_list_from_csv(filepath)
            remove(filepath)
            assert newll == [
                ["h1", "h3", "h4"],
                ["1", "3", "a"],
                ["4", "6", "b"],
                ["7", "9", "c"],
            ]

            with pytest.raises(ValueError):
                read_list_from_csv(filepath, dict_form=True)

    def test_args_to_dict(self):
        args = "a=1,big=hello,1=3"
        assert args_to_dict(args) == {"a": "1", "big": "hello", "1": "3"}



"""Directory Path Utilities."""
import contextlib
import inspect
import logging
import sys
from os import getenv, makedirs, remove
from os.path import (
    abspath,
    basename,
    dirname,
    exists,
    join,
    realpath,
    splitext,
)
from shutil import rmtree
from tempfile import gettempdir
from typing import Any, Dict, Iterable, Optional, Tuple
from urllib.parse import unquote_plus, urlsplit

from slugify import slugify

from .loader import load_text
from .saver import save_text
from .typehint import ListTuple
from .uuid import get_uuid

logger = logging.getLogger(__name__)


class NotFoundError(Exception):
    pass


def script_dir(pyobject: Any, follow_symlinks: bool = True) -> str:
    """Get current script's directory.

    Args:
        pyobject (Any): Any Python object in the script
        follow_symlinks (bool): Follow symlinks or not. Defaults to True.

    Returns:
        str: Current script's directory
    """
    if getattr(sys, "frozen", False):  # py2exe, PyInstaller, cx_Freeze
        path = abspath(sys.executable)  # pragma: no cover
    else:
        path = inspect.getabsfile(pyobject)
    if follow_symlinks:
        path = realpath(path)
    return dirname(path)


def script_dir_plus_file(
    filename: str, pyobject: Any, follow_symlinks: bool = True
) -> str:
    """Get current script's directory and then append a filename.

    Args:
        filename (str): Filename to append to directory path
        pyobject (Any): Any Python object in the script
        follow_symlinks (bool): Follow symlinks or not. Defaults to True.

    Returns:
        str: Current script's directory and with filename appended
    """
    return join(script_dir(pyobject, follow_symlinks), filename)


def get_temp_dir(
    folder: Optional[str] = None,
    delete_if_exists: bool = False,
    tempdir: Optional[str] = None,
) -> str:
    """Get a temporary directory. Looks for environment variable TEMP_DIR and
    falls back on os.gettempdir if a root temporary directory is not supplied.
    If a folder is supplied, creates that folder within the temporary
    directory. Optionally deletes and recreates it if it already exists.

    Args:
        folder (Optional[str]): Folder to create in temporary folder. Defaults to None.
        delete_if_exists (bool): Whether to delete the folder if it exists. Defaults to False.
        tempdir (Optional[str]): Folder to use as temporary directory. Defaults to None (TEMP_DIR or os.gettempdir).

    Returns:
        str: A temporary directory
    """
    if tempdir is None:
        tempdir = getenv("TEMP_DIR", gettempdir())
    if folder:
        tempdir = join(tempdir, folder)
        if exists(tempdir):
            if delete_if_exists:
                rmtree(tempdir)
                makedirs(tempdir)
        else:
            makedirs(tempdir)
    return tempdir


@contextlib.contextmanager
def temp_dir(
    folder: Optional[str] = None,
    delete_if_exists: bool = False,
    delete_on_success: bool = True,
    delete_on_failure: bool = True,
    tempdir: Optional[str] = None,
) -> str:
    """Get a temporary directory optionally with folder appended (and created
    if it doesn't exist)

    Args:
        folder (Optional[str]): Folder to create in temporary folder. Defaults to None.
        delete_if_exists (bool): Whether to delete the folder if it exists. Defaults to False.
        delete_on_success (bool): Whether to delete folder (if folder supplied) on exiting with statement successfully. Defaults to True.
        delete_on_failure (bool): Whether to delete folder (if folder supplied) on exiting with statement unsuccessfully. Defaults to True.
        tempdir (Optional[str]): Folder to use as temporary directory. Defaults to None (TEMP_DIR or os.gettempdir).

    Returns:
        str: A temporary directory
    """
    tempdir = get_temp_dir(
        folder, delete_if_exists=delete_if_exists, tempdir=tempdir
    )
    try:
        yield tempdir
        if folder and delete_on_success:
            rmtree(tempdir)
    except Exception as ex:
        if not isinstance(ex, NotFoundError):
            if folder and delete_on_failure:
                rmtree(tempdir)
            raise


def read_or_create_batch(folder: str, batch: Optional[str] = None) -> str:
    """Get batch or create it if it doesn't exist.

    Args:
        folder (str): Folder in which to look for or create batch file.
        batch (Optional[str]): Batch to use if there isn't one in a file already.

    Returns:
        str: Batch
    """
    batch_file = join(folder, "batch.txt")
    if exists(batch_file):
        batch = load_text(batch_file, strip=True)
        logger.info(f"File BATCH = {batch}")
    else:
        if not batch:
            batch = get_uuid()
            logger.info(f"Generated BATCH = {batch}")
        save_text(batch, batch_file)
    return batch


@contextlib.contextmanager
def temp_dir_batch(
    folder: Optional[str] = None,
    delete_if_exists: bool = False,
    delete_on_success: bool = True,
    delete_on_failure: bool = True,
    batch: Optional[str] = None,
    tempdir: Optional[str] = None,
) -> Dict:
    """Get a temporary directory and batch id. Yields a dictionary with key
    folder which is the temporary directory optionally with folder appended
    (and created if it doesn't exist). In key batch is a batch code to be
    passed as the batch parameter in create_in_hdx or update_in_hdx calls.

    Args:
        folder (Optional[str]): Folder to create in temporary folder. Defaults to None.
        delete_if_exists (bool): Whether to delete the folder if it exists. Defaults to False.
        delete_on_success (bool): Whether to delete folder (if folder supplied) on exiting with statement successfully. Defaults to True.
        delete_on_failure (bool): Whether to delete folder (if folder supplied) on exiting with statement unsuccessfully. Defaults to True.
        batch (Optional[str]): Batch to use if there isn't one in a file already.
        tempdir (Optional[str]): Folder to use as temporary directory. Defaults to None (TEMP_DIR or os.gettempdir).

    Returns:
        Dict: Dictionary containing temporary directory in key folder and batch id in key batch
    """
    with temp_dir(
        folder,
        delete_if_exists,
        delete_on_success,
        delete_on_failure,
        tempdir=tempdir,
    ) as tempdir:
        yield {
            "folder": tempdir,
            "batch": read_or_create_batch(tempdir, batch),
        }


def get_wheretostart(text: str, message: str, key: str) -> Optional[str]:
    """Evaluate WHERETOSTART.

    Args:
        text (str): String to process
        message (str): Text for logging
        key (str): Key to comapre with

    Returns:
        Optional[str]: A string or None
    """
    upper_text = text.upper()
    if upper_text == "RESET":
        return None
    w_key, wheretostart = text.split("=")
    if w_key == key:
        logger.info(f"{message} WHERETOSTART = {wheretostart}")
        return wheretostart
    else:
        return "IGNORE"


def progress_storing_folder(
    info: Dict,
    iterator: Iterable[Dict],
    key: str,
    wheretostart: Optional[str] = None,
) -> Tuple[Dict, Dict]:
    """Store progress in folder in key folder of info dictionary parameter.
    Yields 2 dictionaries. The first is the info dictionary. It contains in key
    folder the folder being used to store progress and in key progress the
    current position in the iterator. If store_batch is True, that dictionary
    will also contain the key batch containing a batch code to be passed as the
    batch parameter in create_in_hdx or update_in_hdx calls. The second
    dictionary is the next dictionary in the iterator.

    Args:
        info (Dict): Dictionary containing folder and anything else to be yielded
        iterator (Iterable[Dict]): Iterate over this object persisting progress
        key (str): Key to examine from dictionary from iterator
        wheretostart (Optional[str]): Where in iterator to start

    Returns:
        Tuple[Dict,Dict]: A tuple of the form (info dictionary, next object in iterator)
    """
    folder = info["folder"]
    progress_file = join(folder, "progress.txt")

    if not wheretostart:
        contents = getenv("WHERETOSTART")
        if contents:
            wheretostart = get_wheretostart(
                contents, "Environment variable", key
            )
        else:
            if exists(progress_file):
                contents = load_text(progress_file, strip=True)
                wheretostart = get_wheretostart(contents, "File", key)
            else:
                wheretostart = None
    found = False
    for nextdict in iterator:
        current = nextdict[key]
        if wheretostart:
            if wheretostart == "IGNORE":
                continue
            if not found:
                if current == wheretostart:
                    found = True
                    logger.info(
                        f"Starting run from WHERETOSTART {wheretostart}"
                    )
                else:
                    logger.info(
                        "Run not started. Ignoring {}. WHERETOSTART ({}) not matched.".format(
                            current, wheretostart
                        )
                    )
                    continue
        output = f"{key}={current}"
        info["progress"] = output
        save_text(output, progress_file)
        yield info, nextdict
    if wheretostart and not found:
        raise NotFoundError(
            f"WHERETOSTART ({wheretostart}) not matched in iterator with key {key} and no run started!"
        )


@contextlib.contextmanager
def wheretostart_tempdir_batch(
    folder: str, batch: Optional[str] = None, tempdir: Optional[str] = None
) -> Dict:
    """Get a temporary directory and batch id. Deletes any existing folder if
    WHERETOSTART environment variable is set to RESET. Yields a dictionary with
    key folder which is the temporary directory optionally with folder appended
    (and created if it doesn't exist). In key batch is a batch code to be
    passed as the batch parameter in create_in_hdx or update_in_hdx calls.

    Args:
        folder (str): Folder to create in temporary folder
        batch (Optional[str]): Batch to use if there isn't one in a file already.
        tempdir (Optional[str]): Folder to use as temporary directory. Defaults to None (TEMP_DIR or os.gettempdir).

    Returns:
        Dict: Dictionary containing temporary directory in key folder and batch id in key batch
    """
    delete_if_exists = False
    wheretostart = getenv("WHERETOSTART")
    if wheretostart:
        if wheretostart.upper() == "RESET":
            delete_if_exists = True
            logger.info(
                "Removing progress file and will start from beginning!"
            )
    with temp_dir_batch(
        folder,
        delete_if_exists,
        delete_on_success=True,
        delete_on_failure=False,
        batch=batch,
        tempdir=tempdir,
    ) as info:
        yield info


def progress_storing_tempdir(
    folder: str,
    iterator: Iterable[Dict],
    key: str,
    batch: Optional[str] = None,
    tempdir: Optional[str] = None,
) -> Tuple[Dict, Dict]:
    """Store progress in temporary directory. The folder persists until the
    final iteration allowing which iteration to start at and the batch code to
    be persisted between runs. Yields 2 dictionaries. The first contains key
    folder which is the temporary directory optionally with folder appended
    (and created if it doesn't exist). In key progress is held the current
    position in the iterator. It also contains the key batch containing a batch
    code to be passed as the batch parameter in create_in_hdx or update_in_hdx
    calls. The second dictionary is the next dictionary in the iterator. The
    WHERETOSTART environment variable can be set to RESET to force the deletion
    and recreation of the temporary directory or to a key value pair in the
    form key=value eg. iso3=PAK indicating where to start.

    Args:
        folder (str): Folder to create in temporary folder
        iterator (Iterable[Dict]): Iterate over the iterator persisting progress
        key (str): Key to examine from dictionary from iterator
        batch (Optional[str]): Batch to use if there isn't one in a file already.
        tempdir (Optional[str]): Folder to use as temporary directory. Defaults to None (TEMP_DIR or os.gettempdir).

    Returns:
        Tuple[Dict,Dict]: A tuple of the form (info dictionary, next object in iterator)
    """
    with wheretostart_tempdir_batch(
        folder, batch=batch, tempdir=tempdir
    ) as info:
        yield from progress_storing_folder(info, iterator, key)


def multiple_progress_storing_tempdir(
    folder: str,
    iterators: ListTuple[Iterable[Dict]],
    keys: ListTuple[str],
    batch: Optional[str] = None,
) -> Tuple[Dict, Dict]:
    """Store progress in temporary directory. The folder persists until the
    final iteration of the last iterator allowing which iteration to start at
    and the batch code to be persisted between runs. Yields 2 dictionaries. The
    first contains key folder which is the temporary directory optionally with
    folder appended (and created if it doesn't exist). In key progress is held
    the current position in the iterator. It also contains the key batch
    containing a batch code to be passed as the batch parameter in
    create_in_hdx or update_in_hdx calls. The second dictionary is the next
    dictionary in the iterator. The WHERETOSTART environment variable can be
    set to RESET to force the deletion and recreation of the temporary
    directory or to a key value pair in the form key=value eg. iso3=PAK
    indicating where to start.

    Args:
        folder (str): Folder to create in temporary folder
        iterators (ListTuple[Iterable[Dict]): Iterate over each iterator in the list consecutively persisting progress
        keys (ListTuple[str]): Key to examine from dictionary from each iterator in the above list
        batch (Optional[str]): Batch to use if there isn't one in a file already.

    Returns:
        Tuple[int, Dict,Dict]: A tuple of the form (iterator index, info dictionary, next object in iterator)
    """
    delete_if_exists = False
    wheretostartenv = getenv("WHERETOSTART")
    if wheretostartenv:
        if wheretostartenv.upper() == "RESET":
            delete_if_exists = True
            logger.info(
                "Removing progress file and will start from beginning!"
            )
    with temp_dir_batch(
        folder,
        delete_if_exists,
        delete_on_success=True,
        delete_on_failure=False,
        batch=batch,
    ) as info:
        tempdir = info["folder"]
        batch = info["batch"]
        for i, key in enumerate(keys):
            progress_file = join(tempdir, "progress.txt")
            if wheretostartenv:
                wheretostart = get_wheretostart(
                    wheretostartenv, "Environment variable", key
                )
            else:
                if exists(progress_file):
                    contents = load_text(progress_file, strip=True)
                    wheretostart = get_wheretostart(contents, "File", key)
                else:
                    wheretostart = None
            with temp_dir_batch(
                str(i),
                False,
                delete_on_success=True,
                delete_on_failure=False,
                batch=batch,
                tempdir=tempdir,
            ) as info:
                for info, nextdict in progress_storing_folder(
                    info, iterators[i], key, wheretostart
                ):
                    save_text(info["progress"], progress_file)
                    yield i, info, nextdict
                if exists(progress_file):
                    remove(progress_file)


def get_filename_extension_from_url(
    url: str, second_last: bool = False, use_query: bool = False
) -> Tuple[str, str]:
    """Get separately filename and extension from url.

    Args:
        url (str): URL to download
        second_last (bool): Get second last segment of url as well. Defaults to False.
        use_query (bool): Include query parameters as well. Defaults to False.

    Returns:
        Tuple[str,str]: Tuple of (filename, extension)
    """
    split_url = urlsplit(unquote_plus(url))
    urlpath = split_url.path
    last_part = basename(urlpath)
    second_last_part = basename(dirname(urlpath))
    query_part = slugify(split_url.query)
    filename, extension = splitext(last_part)
    if query_part:
        if not filename:
            filename = query_part
        elif use_query:
            filename = f"{filename}_{query_part}"
    if second_last_part:
        if not filename:
            filename = second_last_part
        elif second_last:
            filename = f"{second_last_part}_{filename}"
    return filename, extension


def get_filename_from_url(
    url: str, second_last: bool = False, use_query: bool = False
) -> str:
    """Get filename including extension from url.

    Args:
        url (str): URL
        second_last (bool): Get second last segment of url as well. Defaults to False.
        use_query (bool): Include query parameters as well. Defaults to False.

    Returns:
        str: filename
    """
    filename, extension = get_filename_extension_from_url(
        url, second_last, use_query
    )
    return f"{filename}{extension}"



"""State Utility Tests"""
from datetime import datetime, timezone
from os.path import join
from shutil import copyfile

import pytest

from hdx.utilities.dateparse import iso_string_from_datetime, parse_date
from hdx.utilities.path import temp_dir
from hdx.utilities.state import State


class TestState:
    @pytest.fixture(scope="class")
    def statefolder(self, fixturesfolder):
        return join(fixturesfolder, "state")

    @pytest.fixture(scope="class")
    def statefile(self):
        return "last_build_date.txt"

    def test_state(self, statefolder, statefile):
        with temp_dir(folder="test_state") as tmpdir:
            statepath = join(tmpdir, statefile)
            copyfile(join(statefolder, statefile), statepath)
            date1 = datetime(2020, 9, 23, 0, 0, tzinfo=timezone.utc)
            date2 = datetime(2022, 5, 12, 10, 15, tzinfo=timezone.utc)
            with State(
                statepath, parse_date, iso_string_from_datetime
            ) as state:
                assert state.get() == date1
            with State(
                statepath, parse_date, iso_string_from_datetime
            ) as state:
                assert state.get() == date1
                state.set(date2)
            with State(
                statepath, parse_date, iso_string_from_datetime
            ) as state:
                assert state.get() == date2.replace(hour=0, minute=0)



"""Encoding utilities."""

import base64
from typing import Tuple
from urllib.parse import quote, unquote


def str_to_base64(string: str) -> str:
    """Base 64 encode string.

    Args:
        string (str): String to encode

    Returns:
        str: Base 64 encoded string
    """
    return base64.urlsafe_b64encode(string.encode("utf-8")).decode("utf-8")


def base64_to_str(bstring: str) -> str:
    """Base 64 decode string.

    Args:
        bstring (str): Base 64 encoded string to encode

    Returns:
        str: Decoded string
    """
    return base64.urlsafe_b64decode(bstring.encode("utf-8")).decode("utf-8")


def basicauth_encode(username: str, password: str) -> str:
    """Returns an HTTP basic authentication string given a username and
    password.

    Inspired by: https://github.com/rdegges/python-basicauth/blob/master/basicauth.py#L16

    Args:
        username (str): Username
        password (str): Password

    Returns:
        str: Basic authentication string
    """
    if ":" in username:
        raise ValueError

    username_password = f"{quote(username)}:{quote(password)}"
    return "Basic " + str_to_base64(username_password)


def basicauth_decode(encoded_string: str) -> Tuple[str, str]:
    """Decode a HTTP basic authentication string. Returns a tuple of the form
    (username, password), and raises ValueError if decoding fails.

    Inspired by: https://github.com/rdegges/python-basicauth/blob/master/basicauth.py#L27

    Args:
        encoded_string (str): String to decode

    Returns:
        Tuple[str, str]: Tuple of form (username, password)
    """
    split_encoded_string = encoded_string.strip().split(" ")

    if len(split_encoded_string) == 1:
        info_index = 0
    elif (
        len(split_encoded_string) == 2
        and split_encoded_string[0].strip().lower() == "basic"
    ):
        info_index = 1
    else:
        raise ValueError(
            f"Authorization string {encoded_string} should have format "
            f'"xxxxxxxxxxxx" or "Basic xxxxxxxxxxxx"'
        )

    username, password = base64_to_str(split_encoded_string[info_index]).split(
        ":", 1
    )
    return unquote(username), unquote(password)
